@@ -315,7 +315,7 @@ int main(int argc, char *argv[])
         xs *msg = msg_follow(&snac, url);
 
         if (msg != NULL) {
-            char *actor = xs_dict_get(msg, "object");
+            const char *actor = xs_dict_get(msg, "object");
 
             following_add(&snac, actor, msg);
 
@@ -382,7 +382,7 @@ int main(int argc, char *argv[])
         xs *r = content_search(&snac, url, 1, 0, XS_ALL, 10, &to);
 
         int c = 0;
-        char *v;
+        const char *v;
 
         /* print results as standalone links */
         while (xs_list_next(r, &v, &c)) {
