@@ -21,8 +21,8 @@ typedef enum {
     XSTYPE_FALSE  = 0x15,       /* Boolean */
     XSTYPE_LIST   = 0x1d,       /* Sequence of LITEMs up to EOM (with size) */
     XSTYPE_LITEM  = 0x1f,       /* Element of a list (any type) */
-    XSTYPE_DICT   = 0x1c,       /* Sequence of DITEMs up to EOM (with size) */
-    XSTYPE_DITEM  = 0x1e,       /* Element of a dict (STRING key + any type) */
+    XSTYPE_DICT   = 0x1c,       /* Sequence of KEYVALs up to EOM (with size) */
+    XSTYPE_KEYVAL = 0x1e,       /* key + value (STRING key + any type) */
     XSTYPE_EOM    = 0x19,       /* End of Multiple (LIST or DICT) */
     XSTYPE_DATA   = 0x10        /* A block of anonymous data */
 } xstype;
@@ -32,6 +32,7 @@ typedef enum {
 typedef char xs_val;
 typedef char xs_str;
 typedef char xs_list;
+typedef char xs_keyval;
 typedef char xs_dict;
 typedef char xs_number;
 typedef char xs_data;
@@ -45,21 +46,27 @@ typedef char xs_data;
 /* not really all, just very much */
 #define XS_ALL 0xfffffff
 
+#ifndef xs_countof
+#define xs_countof(a) (sizeof((a)) / sizeof((*a)))
+#endif
+
 void *xs_free(void *ptr);
 void *_xs_realloc(void *ptr, size_t size, const char *file, int line, const char *func);
-#define xs_realloc(ptr, size) _xs_realloc(ptr, size, __FILE__, __LINE__, __FUNCTION__)
+#define xs_realloc(ptr, size) _xs_realloc(ptr, size, __FILE__, __LINE__, __func__)
 int _xs_blk_size(int sz);
 void _xs_destroy(char **var);
 #define xs_debug() raise(SIGTRAP)
 xstype xs_type(const xs_val *data);
 int xs_size(const xs_val *data);
 int xs_is_null(const xs_val *data);
+int xs_cmp(const xs_val *v1, const xs_val *v2);
 xs_val *xs_dup(const xs_val *data);
 xs_val *xs_expand(xs_val *data, int offset, int size);
 xs_val *xs_collapse(xs_val *data, int offset, int size);
 xs_val *xs_insert_m(xs_val *data, int offset, const char *mem, int size);
 #define xs_insert(data, offset, data2) xs_insert_m(data, offset, data2, xs_size(data2))
 #define xs_append_m(data, mem, size) xs_insert_m(data, xs_size(data) - 1, mem, size)
+xs_val *xs_stock(int type);
 
 xs_str *xs_str_new(const char *str);
 xs_str *xs_str_new_sz(const char *mem, int sz);
@@ -85,13 +92,14 @@ xs_str *xs_tolower_i(xs_str *str);
 
 xs_list *xs_list_new(void);
 xs_list *xs_list_append_m(xs_list *list, const char *mem, int dsz);
-#define xs_list_append(list, data) xs_list_append_m(list, data, xs_size(data))
-int xs_list_iter(xs_list **list, xs_val **value);
+xs_list *_xs_list_append(xs_list *list, const xs_val *vals[]);
+#define xs_list_append(list, ...) _xs_list_append(list, (const xs_val *[]){ __VA_ARGS__, NULL })
+int xs_list_iter(xs_list **list, const xs_val **value);
+int xs_list_next(const xs_list *list, const xs_val **value, int *ctxt);
 int xs_list_len(const xs_list *list);
-xs_val *xs_list_get(const xs_list *list, int num);
+const xs_val *xs_list_get(const xs_list *list, int num);
 xs_list *xs_list_del(xs_list *list, int num);
 xs_list *xs_list_insert(xs_list *list, int num, const xs_val *data);
-xs_list *xs_list_insert_sorted(xs_list *list, const char *str);
 xs_list *xs_list_set(xs_list *list, int num, const xs_val *data);
 xs_list *xs_list_dequeue(xs_list *list, xs_val **data, int last);
 #define xs_list_pop(list, data) xs_list_dequeue(list, data, 1)
@@ -102,15 +110,20 @@ xs_list *xs_split_n(const char *str, const char *sep, int times);
 #define xs_split(str, sep) xs_split_n(str, sep, XS_ALL)
 xs_list *xs_list_cat(xs_list *l1, const xs_list *l2);
 
+int xs_keyval_size(const xs_str *key, const xs_val *value);
+xs_str *xs_keyval_key(const xs_keyval *keyval);
+xs_val *xs_keyval_value(const xs_keyval *keyval);
+xs_keyval *xs_keyval_make(xs_keyval *keyval, const xs_str *key, const xs_val *value);
+
 xs_dict *xs_dict_new(void);
-xs_dict *xs_dict_append_m(xs_dict *dict, const xs_str *key, const xs_val *mem, int dsz);
-#define xs_dict_append(dict, key, data) xs_dict_append_m(dict, key, data, xs_size(data))
-xs_dict *xs_dict_prepend_m(xs_dict *dict, const xs_str *key, const xs_val *mem, int dsz);
-#define xs_dict_prepend(dict, key, data) xs_dict_prepend_m(dict, key, data, xs_size(data))
-int xs_dict_iter(xs_dict **dict, xs_str **key, xs_val **value);
-xs_val *xs_dict_get(const xs_dict *dict, const xs_str *key);
+xs_dict *xs_dict_append(xs_dict *dict, const xs_str *key, const xs_val *value);
+xs_dict *xs_dict_prepend(xs_dict *dict, const xs_str *key, const xs_val *value);
+int xs_dict_next(const xs_dict *dict, const xs_str **key, const xs_val **value, int *ctxt);
+const xs_val *xs_dict_get_def(const xs_dict *dict, const xs_str *key, const xs_val *def);
+#define xs_dict_get(dict, key) xs_dict_get_def(dict, key, NULL)
 xs_dict *xs_dict_del(xs_dict *dict, const xs_str *key);
 xs_dict *xs_dict_set(xs_dict *dict, const xs_str *key, const xs_val *data);
+xs_dict *xs_dict_gc(xs_dict *dict);
 
 xs_val *xs_val_new(xstype t);
 xs_number *xs_number_new(double f);
@@ -134,27 +147,13 @@ unsigned int xs_hash_func(const char *data, int size);
 #define XS_ASSERT_TYPE_NULL(v, t) (void)(0)
 #endif
 
-extern xs_val xs_stock_null[];
-extern xs_val xs_stock_true[];
-extern xs_val xs_stock_false[];
-extern xs_val xs_stock_0[];
-extern xs_val xs_stock_1[];
-extern xs_val xs_stock_list[];
-extern xs_val xs_stock_dict[];
-
 #define xs_return(v) xs_val *__r = v; v = NULL; return __r
 
+#define xs_is_true(v) (xs_type((v)) == XSTYPE_TRUE)
+#define xs_is_false(v) (xs_type((v)) == XSTYPE_FALSE)
 
 #ifdef XS_IMPLEMENTATION
 
-xs_val xs_stock_null[]  = { XSTYPE_NULL };
-xs_val xs_stock_true[]  = { XSTYPE_TRUE };
-xs_val xs_stock_false[] = { XSTYPE_FALSE };
-xs_val xs_stock_0[]     = { XSTYPE_NUMBER, '0', '\0' };
-xs_val xs_stock_1[]     = { XSTYPE_NUMBER, '1', '\0' };
-xs_val xs_stock_list[]  = { XSTYPE_LIST, 0, 0, 0, 1 + _XS_TYPE_SIZE + 1, XSTYPE_EOM };
-xs_val xs_stock_dict[]  = { XSTYPE_DICT, 0, 0, 0, 1 + _XS_TYPE_SIZE + 1, XSTYPE_EOM };
-
 void *_xs_realloc(void *ptr, size_t size, const char *file, int line, const char *func)
 {
     xs_val *ndata = realloc(ptr, size);
@@ -254,7 +253,7 @@ xstype xs_type(const xs_val *data)
     case XSTYPE_LIST:
     case XSTYPE_LITEM:
     case XSTYPE_DICT:
-    case XSTYPE_DITEM:
+    case XSTYPE_KEYVAL:
     case XSTYPE_NUMBER:
     case XSTYPE_EOM:
     case XSTYPE_DATA:
@@ -272,21 +271,16 @@ xstype xs_type(const xs_val *data)
 void _xs_put_size(xs_val *ptr, int i)
 /* must match _XS_TYPE_SIZE */
 {
-    unsigned char *p = (unsigned char *)ptr;
-
-    p[0] = (i >> 24) & 0x7f;
-    p[1] = (i >> 16) & 0xff;
-    p[2] = (i >> 8) & 0xff;
-    p[3] = i & 0xff;
+    memcpy(ptr + 1, &i, sizeof(i));
 }
 
 
 int _xs_get_size(const xs_val *ptr)
 /* must match _XS_TYPE_SIZE */
 {
-    unsigned char *p = (unsigned char *)ptr;
-
-    return (p[0] << 24) | (p[1] << 16) | (p[2] << 8) | p[3];
+    int i;
+    memcpy(&i, ptr + 1, sizeof(i));
+    return i;
 }
 
 
@@ -307,11 +301,11 @@ int xs_size(const xs_val *data)
     case XSTYPE_LIST:
     case XSTYPE_DICT:
     case XSTYPE_DATA:
-        len = _xs_get_size(data + 1);
+        len = _xs_get_size(data);
 
         break;
 
-    case XSTYPE_DITEM:
+    case XSTYPE_KEYVAL:
         /* calculate the size of the key and the value */
         p = data + 1;
         p += xs_size(p);
@@ -350,13 +344,28 @@ int xs_is_null(const xs_val *data)
 }
 
 
+int xs_cmp(const xs_val *v1, const xs_val *v2)
+/* compares two values */
+{
+    int s1 = xs_size(v1);
+    int s2 = xs_size(v2);
+    int d = s1 - s2;
+
+    return d == 0 ? memcmp(v1, v2, s1) : d;
+}
+
+
 xs_val *xs_dup(const xs_val *data)
 /* creates a duplicate of data */
 {
-    int sz = xs_size(data);
-    xs_val *s = xs_realloc(NULL, _xs_blk_size(sz));
+    xs_val *s = NULL;
 
-    memcpy(s, data, sz);
+    if (data) {
+        int sz = xs_size(data);
+        s = xs_realloc(NULL, _xs_blk_size(sz));
+
+        memcpy(s, data, sz);
+    }
 
     return s;
 }
@@ -380,7 +389,7 @@ xs_val *xs_expand(xs_val *data, int offset, int size)
     if (xs_type(data) == XSTYPE_LIST ||
         xs_type(data) == XSTYPE_DICT ||
         xs_type(data) == XSTYPE_DATA)
-        _xs_put_size(data + 1, sz);
+        _xs_put_size(data, sz);
 
     return data;
 }
@@ -405,7 +414,7 @@ xs_val *xs_collapse(xs_val *data, int offset, int size)
     if (xs_type(data) == XSTYPE_LIST ||
         xs_type(data) == XSTYPE_DICT ||
         xs_type(data) == XSTYPE_DATA)
-        _xs_put_size(data + 1, sz);
+        _xs_put_size(data, sz);
 
     return xs_realloc(data, _xs_blk_size(sz));
 }
@@ -421,6 +430,39 @@ xs_val *xs_insert_m(xs_val *data, int offset, const char *mem, int size)
 }
 
 
+xs_val *xs_stock(int type)
+/* returns stock values */
+{
+    static xs_val stock_null[]  = { XSTYPE_NULL };
+    static xs_val stock_true[]  = { XSTYPE_TRUE };
+    static xs_val stock_false[] = { XSTYPE_FALSE };
+    static xs_val stock_0[]     = { XSTYPE_NUMBER, '0', '\0' };
+    static xs_val stock_1[]     = { XSTYPE_NUMBER, '1', '\0' };
+    static xs_list *stock_list = NULL;
+    static xs_dict *stock_dict = NULL;
+
+    switch (type) {
+    case 0:            return stock_0;
+    case 1:            return stock_1;
+    case XSTYPE_NULL:  return stock_null;
+    case XSTYPE_TRUE:  return stock_true;
+    case XSTYPE_FALSE: return stock_false;
+
+    case XSTYPE_LIST:
+        if (stock_list == NULL)
+            stock_list = xs_list_new();
+        return stock_list;
+
+    case XSTYPE_DICT:
+        if (stock_dict == NULL)
+            stock_dict = xs_dict_new();
+        return stock_dict;
+    }
+
+    return NULL;
+}
+
+
 /** strings **/
 
 xs_str *xs_str_new(const char *str)
@@ -631,10 +673,14 @@ xs_str *xs_tolower_i(xs_str *str)
 xs_list *xs_list_new(void)
 /* creates a new list */
 {
-    return memcpy(
-        xs_realloc(NULL, _xs_blk_size(sizeof(xs_stock_list))),
-        xs_stock_list, sizeof(xs_stock_list)
-    );
+    int sz = 1 + _XS_TYPE_SIZE + 1;
+    xs_list *l = xs_realloc(NULL, sz);
+    memset(l, XSTYPE_EOM, sz);
+
+    l[0] = XSTYPE_LIST;
+    _xs_put_size(l, sz);
+
+    return l;
 }
 
 
@@ -644,8 +690,8 @@ xs_list *_xs_list_write_litem(xs_list *list, int offset, const char *mem, int ds
     XS_ASSERT_TYPE(list, XSTYPE_LIST);
 
     if (mem == NULL) {
-        mem = xs_stock_null;
-        dsz = sizeof(xs_stock_null);
+        mem = xs_stock(XSTYPE_NULL);
+        dsz = xs_size(mem);
     }
 
     list = xs_expand(list, offset, dsz + 1);
@@ -666,7 +712,23 @@ xs_list *xs_list_append_m(xs_list *list, const char *mem, int dsz)
 }
 
 
-int xs_list_iter(xs_list **list, xs_val **value)
+xs_list *_xs_list_append(xs_list *list, const xs_val *vals[])
+/* adds several values to the list */
+{
+    /* special case: if the first argument is NULL, just insert it */
+    if (*vals == NULL)
+        return xs_list_append_m(list, NULL, 0);
+
+    while (*vals) {
+        list = xs_list_append_m(list, *vals, xs_size(*vals));
+        vals++;
+    }
+
+    return list;
+}
+
+
+int xs_list_iter(xs_list **list, const xs_val **value)
 /* iterates a list value */
 {
     int goon = 1;
@@ -697,23 +759,58 @@ int xs_list_iter(xs_list **list, xs_val **value)
 }
 
 
+int xs_list_next(const xs_list *list, const xs_val **value, int *ctxt)
+/* iterates a list, with context */
+{
+    if (xs_type(list) != XSTYPE_LIST)
+        return 0;
+
+    int goon = 1;
+
+    const char *p = list;
+
+    /* skip the start of the list */
+    if (*ctxt == 0)
+        *ctxt = 1 + _XS_TYPE_SIZE;
+
+    p += *ctxt;
+
+    /* an element? */
+    if (xs_type(p) == XSTYPE_LITEM) {
+        p++;
+
+        *value = p;
+
+        p += xs_size(*value);
+    }
+    else {
+        /* end of list */
+        goon = 0;
+    }
+
+    /* update the context */
+    *ctxt = p - list;
+
+    return goon;
+}
+
+
 int xs_list_len(const xs_list *list)
 /* returns the number of elements in the list */
 {
     XS_ASSERT_TYPE_NULL(list, XSTYPE_LIST);
 
-    int c = 0;
-    xs_list *p = (xs_list *)list;
-    xs_val *v;
+    int c = 0, ct = 0;
+    const xs_val *v;
 
-    while (xs_list_iter(&p, &v))
+    while (xs_list_next(list, &v, &ct))
         c++;
 
     return c;
 }
 
 
-xs_val *xs_list_get(const xs_list *list, int num)
+const xs_val *xs_list_get(const xs_list *list, int num)
 /* returns the element #num */
 {
     XS_ASSERT_TYPE(list, XSTYPE_LIST);
@@ -721,11 +818,10 @@ xs_val *xs_list_get(const xs_list *list, int num)
     if (num < 0)
         num = xs_list_len(list) + num;
 
-    int c = 0;
-    xs_list *p = (xs_list *)list;
-    xs_val *v;
+    int c = 0, ct = 0;
+    const xs_val *v;
 
-    while (xs_list_iter(&p, &v)) {
+    while (xs_list_next(list, &v, &ct)) {
         if (c == num)
             return v;
 
@@ -741,7 +837,7 @@ xs_list *xs_list_del(xs_list *list, int num)
 {
     XS_ASSERT_TYPE(list, XSTYPE_LIST);
 
-    xs_val *v;
+    const xs_val *v;
 
     if ((v = xs_list_get(list, num)) != NULL)
         list = xs_collapse(list, v - 1 - list, xs_size(v - 1));
@@ -755,7 +851,7 @@ xs_list *xs_list_insert(xs_list *list, int num, const xs_val *data)
 {
     XS_ASSERT_TYPE(list, XSTYPE_LIST);
 
-    xs_val *v;
+    const xs_val *v;
     int offset;
 
     if ((v = xs_list_get(list, num)) != NULL)
@@ -767,28 +863,6 @@ xs_list *xs_list_insert(xs_list *list, int num, const xs_val *data)
 }
 
 
-xs_list *xs_list_insert_sorted(xs_list *list, const xs_str *str)
-/* inserts a string in the list in its ordered position */
-{
-    XS_ASSERT_TYPE(list, XSTYPE_LIST);
-    XS_ASSERT_TYPE(str, XSTYPE_STRING);
-
-    char *p, *v;
-    int offset = xs_size(list);
-
-    p = list;
-    while (xs_list_iter(&p, &v)) {
-        /* if this element is greater or equal, insert here */
-        if (strcmp(v, str) >= 0) {
-            offset = v - list;
-            break;
-        }
-    }
-
-    return _xs_list_write_litem(list, offset - 1, str, xs_size(str));
-}
-
-
 xs_list *xs_list_set(xs_list *list, int num, const xs_val *data)
 /* sets the element at #num position */
 {
@@ -806,16 +880,16 @@ xs_list *xs_list_dequeue(xs_list *list, xs_val **data, int last)
 {
     XS_ASSERT_TYPE(list, XSTYPE_LIST);
 
-    xs_list *p = list;
-    xs_val *v  = NULL;
+    int ct = 0;
+    const xs_val *v = NULL;
 
     if (!last) {
         /* get the first */
-        xs_list_iter(&p, &v);
+        xs_list_next(list, &v, &ct);
     }
     else {
         /* iterate to the end */
-        while (xs_list_iter(&p, &v));
+        while (xs_list_next(list, &v, &ct));
     }
 
     if (v != NULL) {
@@ -835,11 +909,11 @@ int xs_list_in(const xs_list *list, const xs_val *val)
     XS_ASSERT_TYPE_NULL(list, XSTYPE_LIST);
 
     int n = 0;
-    xs_list *p = (xs_list *)list;
-    xs_val *v;
+    int ct = 0;
+    const xs_val *v;
     int sz = xs_size(val);
 
-    while (xs_list_iter(&p, &v)) {
+    while (xs_list_next(list, &v, &ct)) {
         if (sz == xs_size(v) && memcmp(val, v, sz) == 0)
             return n;
 
@@ -856,13 +930,13 @@ xs_str *xs_join(const xs_list *list, const char *sep)
     XS_ASSERT_TYPE(list, XSTYPE_LIST);
 
     xs_str *s = NULL;
-    xs_list *p = (xs_list *)list;
-    xs_val *v;
+    const xs_val *v;
     int c = 0;
+    int ct = 0;
     int offset = 0;
     int ssz = strlen(sep);
 
-    while (xs_list_iter(&p, &v)) {
+    while (xs_list_next(list, &v, &ct)) {
         /* refuse to join non-string values */
         if (xs_type(v) == XSTYPE_STRING) {
             int sz;
@@ -932,125 +1006,166 @@ xs_list *xs_list_cat(xs_list *l1, const xs_list *l2)
 }
 
 
+/** keyvals **/
+
+int xs_keyval_size(const xs_str *key, const xs_val *value)
+/* returns the needed size for a keyval */
+{
+    return 1 + xs_size(key) + xs_size(value);
+}
+
+
+xs_str *xs_keyval_key(const xs_keyval *keyval)
+/* returns a pointer to the key of the keyval */
+{
+    return (xs_str *)&keyval[1];
+}
+
+
+xs_val *xs_keyval_value(const xs_keyval *keyval)
+/* returns a pointer to the value of the keyval */
+{
+    return (xs_val *)&keyval[1 + xs_size(xs_keyval_key(keyval))];
+}
+
+
+xs_keyval *xs_keyval_make(xs_keyval *keyval, const xs_str *key, const xs_val *value)
+/* builds a keyval into mem (should have enough size) */
+{
+    keyval[0] = XSTYPE_KEYVAL;
+    memcpy(xs_keyval_key(keyval),   key,   xs_size(key));
+    memcpy(xs_keyval_value(keyval), value, xs_size(value));
+
+    return keyval;
+}
+
+
 /** dicts **/
 
+typedef struct {
+    int value_offset;   /* offset to value (from dict start) */
+    int next;           /* next node in sequential search */
+    int child[4];       /* child nodes in hashed search */
+    char key[];         /* C string key */
+} ditem_hdr;
+
+typedef struct {
+    int size;           /* size of full dict (_XS_TYPE_SIZE) */
+    int first;          /* first node for sequential search */
+    int root;           /* root node for hashed search */
+    ditem_hdr ditems[]; /* the ditems */
+} dict_hdr;
+
+
 xs_dict *xs_dict_new(void)
 /* creates a new dict */
 {
-    return memcpy(
-        xs_realloc(NULL, _xs_blk_size(sizeof(xs_stock_dict))),
-        xs_stock_dict, sizeof(xs_stock_dict)
-    );
+    /* size of dict */
+    int sz = 1 + sizeof(dict_hdr);
+
+    xs_dict *d = xs_realloc(NULL, sz);
+    memset(d, '\0', sz);
+
+    d[0] = XSTYPE_DICT;
+    _xs_put_size(d, sz);
+
+    return d;
 }
 
-
-xs_dict *_xs_dict_write_ditem(xs_dict *dict, int offset, const xs_str *key,
-                          const xs_val *data, int dsz)
-/* inserts a memory block into the dict */
+static int *_xs_dict_locate(const xs_dict *dict, const char *key)
+/* locates a ditem */
 {
-    XS_ASSERT_TYPE(dict, XSTYPE_DICT);
-    XS_ASSERT_TYPE(key, XSTYPE_STRING);
+    unsigned int h = xs_hash_func(key, strlen(key));
 
-    if (data == NULL) {
-        data = xs_stock_null;
-        dsz = sizeof(xs_stock_null);
-    }
+    /* start from the root */
+    dict_hdr *dh = (dict_hdr *)(dict + 1);
+    int *off = &dh->root;
 
-    int ksz = xs_size(key);
+    while (*off) {
+        /* pointer to ditem */
+        ditem_hdr *di = (ditem_hdr *)(dict + *off);
 
-    dict = xs_expand(dict, offset, 1 + ksz + dsz);
+        /* pointer to the key */
+        const char *d_key = di->key;
 
-    dict[offset] = XSTYPE_DITEM;
-    memcpy(&dict[offset + 1], key, ksz);
-    memcpy(&dict[offset + 1 + ksz], data, dsz);
-
-    return dict;
-}
-
-
-xs_dict *xs_dict_append_m(xs_dict *dict, const xs_str *key, const xs_val *mem, int dsz)
-/* appends a memory block to the dict */
-{
-    return _xs_dict_write_ditem(dict, xs_size(dict) - 1, key, mem, dsz);
-}
-
-
-xs_dict *xs_dict_prepend_m(xs_dict *dict, const xs_str *key, const xs_val *mem, int dsz)
-/* prepends a memory block to the dict */
-{
-    return _xs_dict_write_ditem(dict, 4, key, mem, dsz);
-}
-
-
-int xs_dict_iter(xs_dict **dict, xs_str **key, xs_val **value)
-/* iterates a dict value */
-{
-    int goon = 1;
-
-    xs_val *p = *dict;
-
-    /* skip the start of the list */
-    if (xs_type(p) == XSTYPE_DICT)
-        p += 1 + _XS_TYPE_SIZE;
-
-    /* an element? */
-    if (xs_type(p) == XSTYPE_DITEM) {
-        p++;
-
-        *key = p;
-        p += xs_size(*key);
-
-        *value = p;
-        p += xs_size(*value);
-    }
-    else {
-        /* end of list */
-        goon = 0;
-    }
-
-    /* store back the pointer */
-    *dict = p;
-
-    return goon;
-}
-
-
-xs_val *xs_dict_get(const xs_dict *dict, const xs_str *key)
-/* returns the value directed by key */
-{
-    XS_ASSERT_TYPE(dict, XSTYPE_DICT);
-    XS_ASSERT_TYPE(key, XSTYPE_STRING);
-
-    xs_dict *p = (xs_dict *)dict;
-    xs_str *k;
-    xs_val *v;
-
-    while (xs_dict_iter(&p, &k, &v)) {
-        if (strcmp(k, key) == 0)
-            return v;
-    }
-
-    return NULL;
-}
-
-
-xs_dict *xs_dict_del(xs_dict *dict, const xs_str *key)
-/* deletes a key */
-{
-    XS_ASSERT_TYPE(dict, XSTYPE_DICT);
-    XS_ASSERT_TYPE(key, XSTYPE_STRING);
-
-    xs_str *k;
-    xs_val *v;
-    xs_dict *p = dict;
-
-    while (xs_dict_iter(&p, &k, &v)) {
-        if (strcmp(k, key) == 0) {
-            /* the address of the item is just behind the key */
-            char *i = k - 1;
-
-            dict = xs_collapse(dict, i - dict, xs_size(i));
+        if (strcmp(key, d_key) == 0)
             break;
+
+        off = &di->child[h >> 30];
+        h <<= 2;
+    }
+
+    return off;
+}
+
+
+xs_dict *xs_dict_set(xs_dict *dict, const xs_str *key, const xs_val *value)
+/* sets a key/value pair */
+{
+    if (value == NULL)
+        value = xs_stock(XSTYPE_NULL);
+
+    if (xs_type(dict) == XSTYPE_DICT) {
+        int *o = _xs_dict_locate(dict, key);
+        int end = xs_size(dict);
+
+        if (!*o) {
+            /* ditem does not exist yet: append to the end */
+            *o = end;
+
+            int ksz = xs_size(key);
+            int vsz = xs_size(value);
+            int dsz = sizeof(ditem_hdr) + ksz + vsz;
+
+            /* open room in the dict for the full ditem */
+            dict = xs_expand(dict, end, dsz);
+
+            dict_hdr *dh = (dict_hdr *)(dict + 1);
+
+            /* build the ditem */
+            ditem_hdr *di = (ditem_hdr *)(dict + end);
+            memset(di, '\0', dsz);
+
+            /* set the offset to the value */
+            di->value_offset = end + sizeof(ditem_hdr) + ksz;
+
+            /* copy the key */
+            memcpy(di->key, key, ksz);
+
+            /* copy the value */
+            memcpy(dict + di->value_offset, value, vsz);
+
+            /* chain to the sequential list */
+            di->next = dh->first;
+            dh->first = end;
+        }
+        else {
+            /* ditem already exists */
+            ditem_hdr *di = (ditem_hdr *)(dict + *o);
+
+            /* get pointer to the value offset */
+            int *i = &di->value_offset;
+
+            /* deleted? recover offset */
+            if (*i < 0)
+                *i *= -1;
+
+            /* get old value */
+            xs_val *o_value = dict + *i;
+
+            /* will new value fit over the old one? */
+            if (xs_size(value) <= xs_size(o_value)) {
+                /* just overwrite */
+                /* (difference is leaked inside the dict) */
+                memcpy(o_value, value, xs_size(value));
+            }
+            else {
+                /* not enough room: new value will live at the end of the dict */
+                /* (old value is leaked inside the dict) */
+                *i = end;
+
+                dict = xs_insert(dict, end, value);
+            }
         }
     }
 
@@ -1058,22 +1173,109 @@ xs_dict *xs_dict_del(xs_dict *dict, const xs_str *key)
 }
 
 
-xs_dict *xs_dict_set(xs_dict *dict, const xs_str *key, const xs_val *data)
-/* sets (replaces) a key */
+xs_dict *xs_dict_append(xs_dict *dict, const xs_str *key, const xs_val *value)
+/* just an alias (for this implementation it's the same) */
 {
-    XS_ASSERT_TYPE(dict, XSTYPE_DICT);
-    XS_ASSERT_TYPE(key, XSTYPE_STRING);
+    return xs_dict_set(dict, key, value);
+}
 
-    /* delete the possibly existing key */
-    dict = xs_dict_del(dict, key);
 
-    /* add the data */
-    dict = xs_dict_append(dict, key, data);
+xs_dict *xs_dict_prepend(xs_dict *dict, const xs_str *key, const xs_val *value)
+/* just an alias (for this implementation it's the same) */
+{
+    return xs_dict_set(dict, key, value);
+}
+
+
+xs_dict *xs_dict_del(xs_dict *dict, const xs_str *key)
+/* deletes a key/value pair */
+{
+    if (xs_type(dict) == XSTYPE_DICT) {
+        int *o = _xs_dict_locate(dict, key);
+
+        if (*o) {
+            /* found ditem */
+            ditem_hdr *di = (ditem_hdr *)(dict + *o);
+
+            /* deleted ditems have a negative value offset */
+            di->value_offset *= -1;
+        }
+    }
 
     return dict;
 }
 
 
+const xs_val *xs_dict_get_def(const xs_dict *dict, const xs_str *key, const xs_str *def)
+/* gets a value by key, or returns def */
+{
+    if (xs_type(dict) == XSTYPE_DICT) {
+        int *o = _xs_dict_locate(dict, key);
+
+        if (*o) {
+            /* found ditem */
+            ditem_hdr *di = (ditem_hdr *)(dict + *o);
+
+            if (di->value_offset > 0)
+                return dict + di->value_offset;
+        }
+    }
+
+    return def;
+}
+
+
+int xs_dict_next(const xs_dict *dict, const xs_str **key, const xs_val **value, int *ctxt)
+/* dict iterator, with context */
+{
+    if (xs_type(dict) != XSTYPE_DICT)
+        return 0;
+
+    if (*ctxt == 0) {
+        /* at the beginning: get the first sequential item */
+        const dict_hdr *dh = (dict_hdr *)(dict + 1);
+        *ctxt = dh->first;
+    }
+
+    *value = NULL;
+
+    while (*value == NULL && *ctxt > 0) {
+        const ditem_hdr *di = (ditem_hdr *)(dict + *ctxt);
+
+        /* get value */
+        if (di->value_offset > 0) {
+            *value = (xs_val *)(dict + di->value_offset);
+
+            /* get key */
+            *key = (xs_str *)&di->key;
+        }
+
+        /* get offset to next ditem */
+        *ctxt = di->next ? di->next : -1;
+    }
+
+    return *value != NULL;
+}
+
+
+xs_dict *xs_dict_gc(xs_dict *dict)
+/* collects garbage (leaked values) inside a dict */
+{
+    xs_dict *nd = xs_dict_new();
+    const xs_str *k;
+    const xs_val *v;
+    int c = 0;
+
+    /* shamelessly create a new dict with the same content */
+    while (xs_dict_next(dict, &k, &v, &c))
+        nd = xs_dict_set(nd, k, v);
+
+    xs_free(dict);
+
+    return nd;
+}
+
+
 /** other values **/
 
 xs_val *xs_val_new(xstype t)
@@ -1124,8 +1326,11 @@ double xs_number_get(const xs_number *v)
 {
     double f = 0.0;
 
-    if (v != NULL && v[0] == XSTYPE_NUMBER)
+    if (xs_type(v) == XSTYPE_NUMBER)
         f = atof(&v[1]);
+    else
+    if (xs_type(v) == XSTYPE_STRING)
+        f = atof(v);
 
     return f;
 }
@@ -1136,7 +1341,7 @@ const char *xs_number_str(const xs_number *v)
 {
     const char *p = NULL;
 
-    if (v != NULL && v[0] == XSTYPE_NUMBER)
+    if (xs_type(v) == XSTYPE_NUMBER)
         p = &v[1];
 
     return p;
@@ -1156,7 +1361,7 @@ xs_data *xs_data_new(const void *data, int size)
     v = xs_realloc(NULL, _xs_blk_size(total_size));
     v[0] = XSTYPE_DATA;
 
-    _xs_put_size(v + 1, total_size);
+    _xs_put_size(v, total_size);
 
     memcpy(&v[1 + _XS_TYPE_SIZE], data, size);
 
@@ -1167,7 +1372,7 @@ xs_data *xs_data_new(const void *data, int size)
 int xs_data_size(const xs_data *value)
 /* returns the size of the data stored inside value */
 {
-    return _xs_get_size(value + 1) - (1 + _XS_TYPE_SIZE);
+    return _xs_get_size(value) - (1 + _XS_TYPE_SIZE);
 }
 
 
@@ -1203,7 +1408,7 @@ unsigned int xs_hash_func(const char *data, int size)
     int n;
 
     for (n = 0; n < size; n++) {
-        hash ^= data[n];
+        hash ^= (unsigned char)data[n];
         hash *= 111111111;
     }
 
